unit UnitMineUtils;

interface
uses sysutils;
Const
  //Titre de la fentre du dmineur
  WinMineWindowTitle='Dmineur';

Type
  //versions du dmineur
  TWinMineVersion=(v9x,vNT,vXP);
  TMatrice=array of array of Integer;
  TTableauEntiers=array of Integer;
  TTableauReel=array of Extended;

function GetWinMineVersion:TWinMineVersion;

//C(n,p) ==> combinaison de p elements parmis m
function Comb(n,p:Integer):Integer;
//Nbre de fois ou un lment apparait dans C(n,p) combinaisons
function Occurences(n,p:Integer):Integer;

//Rsolution par Satisfaction de contraintes d'un champ de mines
procedure CSPMine(Contraintes:TMatrice;Probas:TTableauReel;NbreNondevinables:integer);
implementation
uses UnitSystem,dialogs;

function GetWinMineVersion:TWinMineVersion;
begin
  Result:=vXP;
  case WindowsVersion of
    Win_NT_351,Win_NT_40:Result:=vNT;
    Win_95,Win_98,Win_ME:Result:=v9x;
    Win_2000,Win_XP,Win_Server_2003:Result:=vXP;
  end;
end;

//!n/!(n-p)=n*(n-1)*...*(n-p+1)
function factoriellenp(n,p:integer):integer;
begin
  if p<=0
  then factoriellenp:=1
  else factoriellenp:=n*factoriellenp(n-1,p-1);
end;

//!n
function factorielle(n:integer):integer;
begin
  if n<=0
  then factorielle:=1
  else factorielle:=n*factorielle(n-1);
end;

//C(n,p) ==> combinaisons/Permutations de p elements parmis n
function Comb(n,p:Integer):integer;
begin
 if (p<0) or (p>n)
 then Result:=0
 else if (p=0) or (p=n)
 then Result:=1
 else Result:=Comb(n-1,p-1)*n div p;
end;

function Comb_r(n,p:Integer):real;
begin
 if (p<0) or (p>n)
 then Result:=0
 else if (p=0) or (p=n)
 then Result:=1
 else Result:=Comb_r(n-1,p-1)*(n/p);
end;

//Nbre de fois ou un lment apparait dans C(n,p) combinaisons
function Occurences(n,p:Integer):Integer;
begin
  result:=Comb(n-1,p-1);
end;

procedure CalculeRang(Matrice:TMatrice;rangs:TTableauEntiers);
var i,j,rj:Integer;
    iCount,jCount:Integer;
begin
  //Tri de la matrice
  iCount:=Length(Matrice);
  jCount:=Length(Matrice[0]);
  for j:=0 to jCount-1 do
  begin
    rangs[j]:=0;
    rj:=-1;
    i:=0;
    while ((i<iCount) and (rj=-1)) do
    begin
      if (Matrice[i,j]<>0) then rj:=i;
      Inc(i);
    end;
    if rj=-1 then rj:=i;
    rangs[j]:=rj;
  end;
end;

procedure CalculeNbreInconnues(Matrice:TMatrice;NbreInconnues:TTableauEntiers);
var i,j,nj:Integer;
    iCount,jCount,NbreValeurs:Integer;
begin
  //Tri de la matrice
  iCount:=Length(Matrice);
  jCount:=Length(Matrice[0]);
  NbreValeurs:=iCount-1;
  for j:=0 to jCount-1 do
  begin
    nj:=0;
    for i:=0 to NbreValeurs-1 do if Matrice[i,j]<>0 then Inc(nj);
    NbreInconnues[j]:=NJ;
  end;
end;

procedure TriLignes(Matrice:TMatrice;rangs:TTableauEntiers);
var i,j,j2,T:Integer;
    iCount,jCount:Integer;
    NbreInconnues:TTableauEntiers;
begin
//Tri de la matrice
  iCount:=Length(Matrice);
  jCount:=Length(Matrice[0]);
  SetLength(NbreInconnues,jCount);

  CalculeRang(Matrice,rangs);
  CalculeNbreInconnues(Matrice,NbreInconnues);

  for j:=0 to jCount-2 do
  begin
    for j2 := (jCount-1) downto (j+1) do
    begin
      if rangs[j]>rangs[j2] then
      begin
        T:=rangs[j];
        rangs[j]:=rangs[j2];
        rangs[j2]:=T;
        for i:=0 to iCount-1 do
        begin
          T:=Matrice[i,j];
          Matrice[i,j]:=Matrice[i,j2];
          Matrice[i,j2]:=T;
        end;
      end;
    end;
  end;
  SetLength(NbreInconnues,0);
end;

function PGCD(a,b:Integer):integer;
var u,v:integer;
begin
 u:=abs(a);
 v:=abs(b);

 while ((u<>0) and (v<>0)) DO
   if u>v then u:=u-v else v:=v-u;
 If u=0 then result:=v else result:=u;
end;

procedure SimplifierCoefs(Matrice:TMatrice);
var i,j,k:Integer;
    iCount,jCount:Integer;
    Diviseur:integer;
    rangs:TTableauEntiers;
begin
  iCount:=Length(Matrice);
  jCount:=Length(Matrice[0]);
  SetLength(rangs,jCount);
  for j:=0 to jCount-1 do
  begin
    Diviseur:=0;
    k:=0;
    for i:=0 to iCount-1 do
    begin
      Diviseur:=PGCD(Diviseur,Matrice[i,j]);
      if (k=0) then k:=Matrice[i,j]
    end;
    if k<0 then Diviseur:=-Diviseur;
    if (Diviseur>1) or (Diviseur<0) then
    for i:=0 to iCount-1 do
      Matrice[i,j]:=Matrice[i,j] div Diviseur;
  end;
  SetLength(rangs,0);
end;



procedure SolveEqs(Equations:TMatrice);
var iCount,jCount,NbreValeurs:Integer;
    i,j,rj,j2:Integer;
    coef1,coef2:integer;
    rangs:TTableauEntiers;
Begin
 iCount:=Length(Equations);
 jCount:=Length(Equations[0]);
 NbreValeurs:=iCount-1;
 SetLength(rangs,jCount);
 TriLignes(Equations,rangs);
 CalculeRang(Equations,rangs);
 //"Rduction" de la matrice par substitution inverse
 for j:=(jCount-1) downto 1 do
 begin
   if rangs[j]<NbreValeurs then
   begin
     coef1:=Equations[rangs[j],j];
     for j2:=(j-1) downto 0 do
     begin
       coef2:=Equations[rangs[j],j2];
       if coef2<>0 then
       //on rduit le nombre d'inconnues de la ligne par substitution
       for i:=0 to iCount-1 do
           Equations[i,j2]:=Equations[i,j2]*coef1-coef2*Equations[i,j];
     end;
   end;
 end;
 SimplifierCoefs(Equations);
 TriLignes(Equations,rangs);
 CalculeRang(Equations,rangs);
 //"triangulation" de la matrice
 for j:=0 to jCount-2 do
 begin
   rj:=rangs[j];
   if rj<NbreValeurs then
   begin
     coef1:=Equations[rj,j];
     for j2:=j+1 to jCount-1 do
     begin
       coef2:=Equations[rj,j2];
       if coef2<>0 then
       for i:=0 to iCount-1 do
           Equations[i,j2]:=Equations[i,j2]*coef1-coef2*Equations[i,j];
       CalculeRang(Equations,rangs);
     end;
   end;
 end;

 TriLignes(Equations,rangs);
 SimplifierCoefs(Equations);
 //
 TriLignes(Equations,rangs);

 //"Rduction" de la matrice par substitution inverse
 for j:=(jCount-1) downto 1 do
 begin
   rj:=rangs[j];
   if rj<NbreValeurs then
   begin
     coef1:=Equations[rj,j];
     for j2:=(j-1) downto 0 do
     begin
       coef2:=Equations[rj,j2];
       if coef2<>0 then
       //on rduit le nombre d'inconnues de la ligne par substitution
       for i:=0 to iCount-1 do
           Equations[i,j2]:=Equations[i,j2]*coef1-coef2*Equations[i,j];
     end;
   end;
 end;

 SimplifierCoefs(Equations);
 SetLength(rangs,0);
end;

//Vrification des domaines des valeurs
//qui doivent tre compris entre 0 et 1
function SolutionEstValide(Solution:TMatrice;NbreNondevinables:integer):Boolean;
var iCount,jCount,NbreValeurs:Integer;
    i,j:integer;
    NbreInconnues:TTableauEntiers;
    SommeCoefsPos:Integer;
    SommeCoefsNeg:Integer;
    ACoef:Integer;
begin
  iCount:=Length(Solution);
  jCount:=Length(Solution[0]);
  NbreValeurs:=iCount-1;
  SetLength(NbreInconnues,jCount);
  CalculeNbreInconnues(Solution,NbreInconnues);
  result:=True;
  j:=0;
  while ((j<jCount) and Result) do
  begin
    SommeCoefsPos:=0;
    SommeCoefsNeg:=0;
    ACoef:=Solution[0,j]*NbreNondevinables;
    if ACoef>0 then SommeCoefsPos:=ACoef else SommeCoefsNeg:=ACoef;
    if NbreValeurs>1 then
      for i:=1 to NbreValeurs-1 do
      begin
        ACoef:=Solution[i,j];
        if ACoef>0
        then SommeCoefsPos:=SommeCoefsPos+ACoef
        else SommeCoefsNeg:=SommeCoefsNeg+ACoef;
      end;
    ACoef:=Solution[iCount-1,j];
    if ACoef>0
    then result:=result and (ACoef<=SommeCoefsPos)
    else result:=result and (ACoef>=SommeCoefsNeg);
    inc(j);
  end;
  SetLength(NbreInconnues,0);
end;

function f(NN,x,xm:Integer):Extended;
var a,b,aa,bb:integer;
    i,j:integer;
begin
  result:=1;
  j:=xm;
  for i:=(NN-x) to (NN-xm) do
  begin
    a:=i;
    b:=j;
    if i<=0 then a:=1;
    if j<=0 then b:=1;
    Result:=Result*a/b;
    dec(j);
  end;
end;

procedure CalculeStats(UneSolution:TMatrice;Probas:TTableauReel;NbreNondevinables:integer;var CardinaliteSolution:Extended;Xr,XrMax:integer);
var iCount,jCount,NbreValeurs:Integer;
    i,j:integer;
    N,P:Extended;
    NbreInconnues:TTableauEntiers;
    VM:Real;
    Probas2:TTableauReel;
begin
  iCount:=Length(UneSolution);
  jCount:=Length(UneSolution[0]);
  NbreValeurs:=iCount-1;
  SetLength(Probas2,NbreValeurs);
  SetLength(NbreInconnues,jCount);
  CalculeNbreInconnues(UneSolution,NbreInconnues);
  //A la place de ceci : Comb_r(NbreNondevinables,UneSolution[max_i,0]);
  //on calcule ((NbreNondevinables-Xr)!/(NbreNondevinables-XMax)!)/(!XrMax/!Xr)!
  //pour viter les grands nombres
  N:=f(NbreNondevinables,Xr,XrMax);
  P:=UneSolution[iCount-1,0];
  Probas2[0]:=P;
  for j:=1 to jCount-1 do
  begin
    if NbreInconnues[j]>0 then
    begin
      VM:=UneSolution[iCount-1,j]/NbreInconnues[j];
      if VM>0 then
      for i:=1 to NbreValeurs-1 do
      begin
        if UneSolution[i,j]<>0 then
           Probas2[i]:=VM;
      end;
      N:=N*comb(NbreInconnues[j],UneSolution[iCount-1,j]);
    end;
  end;
  CardinaliteSolution:=N;
  for i:=0 to NbreValeurs-1 do
  begin
    P:=Probas2[i];
    Probas[i]:=Probas[i]+P*N;
  end;
  SetLength(NbreInconnues,0);
end;

//Rsolution --> module principal
procedure CSPMine(Contraintes:TMatrice;Probas:TTableauReel;NbreNondevinables:integer);
var iCount,jCount,jCount2,NbreValeurs:Integer;
    UneSolution:TMatrice;
    NbreInconnues:TTableauEntiers;
    rangs:TTableauEntiers;

    Xr,{XrMin,}XrMax,NbreSolutionsATester:Integer;
    i,j,k,ii,kk:integer;
    NTot,Ni,P:Extended;
    Dependances:array of integer;
    NbreDependances:Integer;
    IndexOfDependances:array of Integer;
    ValeursDesDependances:Cardinal;

begin
  iCount:=Length(Contraintes);
  if iCount>=0 then jCount:=Length(Contraintes[0]) else exit;
  NbreValeurs:=iCount-1;
  SetLength(Dependances,NbreValeurs);
  SetLength(NbreInconnues,jCount);
  SetLength(rangs,jCount);

  SolveEqs(Contraintes);
  CalculeNbreInconnues(Contraintes,NbreInconnues);
  CalculeRang(Contraintes,rangs);
  NbreDependances:=0;
  for i:=0 to NbreValeurs-1 do Dependances[i]:=-2;
  if NbreValeurs>0 then
    for i:=1 to NbreValeurs-1 do
    begin
      k:=0;
      for j:=0 to jCount-1 do
          if Contraintes[i,j]<>0 then begin
                                        Inc(k);
                                        if j=0 then Inc(k);
                                      end;
      Dependances[i]:=k;
      if k>1 then Inc(NbreDependances);
    end;
  j:=0;
  SetLength(IndexOfDependances,NbreDependances);
  jCount2:=jCount+NbreDependances;


  if NbreDependances>0 then
    for i:=0 to NbreValeurs-1 do
      if Dependances[i]>1
      then begin
             IndexOfDependances[j]:=i;
             Inc(j);
           end;
  NbreSolutionsATester:=(1 shl NbreDependances);

  XrMax:=Contraintes[iCount-1,0];
  //XrMin:=XrMax-NbreInconnues[0]+1;

  Ntot:=0;

  Probas[0]:=0.1;
  for i:=1 to NbreValeurs-1 do Probas[i]:=0.2;
  if NbreSolutionsATester>32 then exit;
  for i:=0 to NbreValeurs-1 do Probas[i]:=0;
  
  if NbreSolutionsATester>0 then
  begin
    for ValeursDesDependances:=0 to NbreSolutionsATester-1 do

    begin
      //Copie de la rduction de contraintes
      SetLength(UneSolution,iCount,jCount);
      UneSolution:=Copy(Contraintes);
      SetLength(UneSolution,iCount,jCount2);
      //Trouve une solution particulire si on a plusieurs solutions
      //en rajoutant quelques quations
      if NbreDependances>0 then
        for kk:=0 to NbreDependances-1 do
        begin
          ii:=IndexOfDependances[kk];
          UneSolution[ii,jCount2-1-kk]:=1;
          if TestBit(ValeursDesDependances,kk) then UneSolution[iCount-1,jCount2-1-kk]:=1
                                               else UneSolution[iCount-1,jCount2-1-kk]:=0;
        end;
      SolveEqs(UneSolution);
      //on tient peut tre une solution valide
      //on met les stats  jours pour cette solution
      Xr:=UneSolution[iCount-1,0];
      if SolutionEstValide(UneSolution,NbreNondevinables) then
      begin
        CalculeStats(UneSolution,Probas,NbreNondevinables,Ni,Xr,XrMax);
        Ntot:=NTot+Ni;
      end;
    end;
    //on calcule les valeurs de proba dfinitives;
    if NTot>0 then
    for i:=0 to NbreValeurs-1 do
    begin
      P:=Probas[i];
      P:=P/NTot;
      Probas[i]:=P;
    end;
    if NbreNondevinables>0 then
    begin
      P:=Probas[0];
      Probas[0]:=P/NbreNondevinables;
    end;

  end;
  SetLength(NbreInconnues,0);
  SetLength(UneSolution,0,0);
  SetLength(Dependances,0);
  SetLength(rangs,0);
end;

end.
