unit UnitStats;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls,UnitCommon;

type
  TStats=record
    TempsTotal,TempsMin,TempsMax:cardinal;
    TempsMoyen:Real;
    Echantillon:array of Cardinal;
  end;

  TStatistiques=class
  private
    //TickcCount du systme pour le chrono du BOT
    FStartTick:Cardinal;
    FStopTick:Cardinal;
    FStatsDemineur: TStats;
    FStatsBot: TStats;
    FNbreTemps: integer;
    FAfficheOfficiel: boolean;
    FNBredeParties: integer;
    procedure SetAfficheOfficiel(const Value: boolean);
    procedure SetNBredeParties(const Value: integer);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Initialiser;
    procedure AjouteTemps(Resultat:TEtatduJeu;Temps:cardinal);
    procedure StartChrono;
    procedure StopChrono;
    property NBredeParties:integer read FNBredeParties write SetNBredeParties;
    property NbreTemps:integer read FNbreTemps;
    //Statistiques d'aprs WinMine
    property StatsDemineur:TStats read FStatsDemineur; //units : secondes
    //d'aprs le BOT
    property StatsBot:TStats read FStatsBot;//units : millisecondes
    property AfficheOfficiel:boolean read FAfficheOfficiel write SetAfficheOfficiel;
  end;

  TFormStats = class(TForm)
    MemoStats: TMemo;
    Button1: TButton;
    StaticText1: TStaticText;
    StaticTextDetail: TStaticText;
    Splitter1: TSplitter;
    ListViewDetails: TListView;
    Panel1: TPanel;
    Panel2: TPanel;
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
  private
    { Dclarations prives }
    oldBtnProc : TWndMethod;
    FNiveau: string;
    Procedure MyBtnWndProc( Var msg : TMessage );
    procedure SetNiveau(const Value: string);
  public
    { Dclarations publiques }
    procedure AfficherStats;
    property Niveau:string read FNiveau write SetNiveau;
  end;

var
  FormStats: TFormStats;
   Statistiques:TStatistiques;

implementation

{$R *.dfm}

procedure TFormStats.FormCreate(Sender: TObject);
begin
  MemoStats.Clear;
  // Sauvegarde la WndProc actuelle du bouton
   oldBtnProc := Button1.WindowProc;
   // Affecte une nouvelle procdure de fentre.
   Button1.WindowProc := MyBtnWndProc;
end;

procedure TFormStats.Button1Click(Sender: TObject);
begin
  Close;
end;

procedure TFormStats.MyBtnWndProc(var msg: TMessage);
Var  btnCanvas : TControlCanvas;
     i,j:Integer;
     AColor,OldColor:TColor;
 Begin
   OldBtnProc(msg);
   If (msg.Msg = BM_SETSTATE) Or (msg.Msg = WM_PAINT) Or (msg.Msg = WM_NCPAINT) Then
   Begin
     btnCanvas := TControlCanvas.Create;  
     Try
       btnCanvas.Control:=Button1;
       OldColor:=TColor(GetTextColor(btnCanvas.Handle)); 
       for i:=2 to Button1.ClientWidth-3 do
         for j:=2 to Button1.ClientHeight-3 do
         begin
           AColor:=btnCanvas.Pixels[i,j];
           if AColor<>OldColor
           then btnCanvas.Pixels[i,j]:=clyellow;
         end;
     Finally
       btnCanvas.Free;
     End;
   End;
 End;

{ TStatistiques }

procedure TStatistiques.AjouteTemps(Resultat:TEtatduJeu;Temps:cardinal);
var TempsBOT:Cardinal;
    ListItem:TListItem;
begin
  if Resultat=edjGagne
  then Inc(FNBreTemps);
  Inc(FNBredeParties);
  ListItem:=FormStats.ListViewDetails.Items.Add;
  ListItem.Caption:=IntToStr(FNBredeParties);
  ListItem.SubItems.Add(MessagesStatut[Resultat]);
  ListItem.SubItems.Add('-');
  ListItem.SubItems.Add('-');
  if Resultat=edjGagne then
  begin
    if AfficheOfficiel then
    with FStatsDemineur do
    begin
      Inc(TempsTotal,Temps);
      SetLength(Echantillon,FNbreTemps);
      Echantillon[FNbreTemps-1]:=Temps;
      if Temps>TempsMax then TempsMax:=Temps;
      if Temps<TempsMin then TempsMin:=Temps;
      TempsMoyen:=TempsTotal/FNbreTemps;
      ListItem.SubItems.Strings[1]:=Format('%u',[Temps]);
    end;
    TempsBOT:=FStopTick-FStartTick;
    with FStatsBot do
    begin
      Inc(TempsTotal,TempsBOT);
      SetLength(Echantillon,FNbreTemps);
      Echantillon[FNbreTemps-1]:=TempsBOT;
      if TempsBOT>TempsMax then TempsMax:=TempsBOT;
      if TempsBOT<TempsMin then TempsMin:=TempsBOT;
      TempsMoyen:=TempsTotal/FNbreTemps;
      ListItem.SubItems.Strings[2]:=Format('%g',[TempsBOT/1000]);
    end;
  end;
  FormStats.AfficherStats;
end;

constructor TStatistiques.Create;
begin
  SetLength(FStatsDemineur.Echantillon,0);
  SetLength(FStatsBot.Echantillon,0);
end;

destructor TStatistiques.Destroy;
begin
  SetLength(FStatsDemineur.Echantillon,0);
  SetLength(FStatsBot.Echantillon,0);
end;

procedure TStatistiques.Initialiser;
begin
  FNBreTemps:=0;
  FNBredeParties:=0;
  with FStatsDemineur do
  begin
    TempsTotal:=0;
    TempsMax:=0;
    TempsMin:=1000;
    TempsMoyen:=0;
    SetLength(Echantillon,0);
  end;
  with FStatsBot do
  begin
    TempsTotal:=0;
    TempsMax:=0;
    TempsMin:=1000000;
    TempsMoyen:=0;
    SetLength(Echantillon,0);
  end;

  FormStats.MemoStats.Clear;
  FormStats.ListViewDetails.Clear;
end;

procedure TStatistiques.SetAfficheOfficiel(const Value: boolean);
begin
  FAfficheOfficiel := Value;
end;

procedure TStatistiques.SetNBredeParties(const Value: integer);
begin
  FNBredeParties := Value;
end;

procedure TStatistiques.StartChrono;
begin
  FStartTick:=GetTickCount;
end;

procedure TStatistiques.StopChrono;
begin
  FStopTick:=GetTickCount;
end;

procedure TFormStats.AfficherStats;
var AfficheStats:string;
begin
  with Statistiques do
  begin
    AfficheStats:='Statistiques :'+#13#10#13#10+
                   'Niveau : '+FNiveau+#13#10#13#10+
                   Format('Parties joues. : %u',[NbreDeParties])+#13#10+
                   Format('Parties gagnes : %u',[NbreTemps])+#13#10+
                   '%'+Format(' gagnes...... : %g',[100*NbreTemps/NbreDeParties])+'%'+#13#10+#13#10;
    if NbreTemps>0
    then begin
           if AfficheOfficiel then
           with StatsDemineur do
           AfficheStats:=AfficheStats+'Temps "Officiel" (selon le dmineur)'+#13#10+
           Format('  Temps moyen lorsque gain : %g secondes',[TempsTotal/NbreTemps])+#13#10+
           Format('  Temps Mini.... : %u seconde(s)',[TempsMin])+#13#10+
           Format('  Temps Maxi.... : %u seconde(s)',[TempsMax])+#13#10;

           with StatsBot do
           AfficheStats:=AfficheStats+'Temps "rel" (calcul par le bot)'+#13#10+
           Format('  Temps moyen lorsque gain : %g secondes',[TempsTotal/NbreTemps/1000])+#13#10+
           Format('  Temps Mini.... : %g seconde(s)',[TempsMin/1000])+#13#10+
           Format('  Temps Maxi.... : %g seconde(s)',[TempsMax/1000]);
         end;
  end;
   MemoStats.Clear;
   MemoStats.Lines.Add(AfficheStats);
end;

procedure TFormStats.SetNiveau(const Value: string);
begin
  FNiveau := Value;
end;

initialization
  Statistiques:=TStatistiques.Create;
finalization
  Statistiques.Free;
end.
